"""
Material parameters for hexagonal boron nitride
"""
import numpy as np
import warnings
from common import *

lightspeed = 299792458. #m/s
cm_1 = 2*np.pi*lightspeed*1e2   # angular frequency from cm^-1

def permittivity_Geick(omega):
    """
    Epsilon of hexagonal boron nitride/epsilon_0.
    This is a two-component permittivity for in-plane electric field,
    out-of-plane electric field.

    This is based on Geick et al., 1966.
    Note that this BN is likely a fairly dirty sample with misaligned
    crystallites. It should not be used for exfoliated monocrystals of
    h-BN.
    """

    ## FROM GEICK (1966)
    perp = (4.95
            +(1.23e5/ 767.**2)*lor(omega, 767.*cm_1,35.*cm_1) #should be inactive
            +(3.49e6/1367.**2)*lor(omega,1367.*cm_1,29.*cm_1)
            )
    par  = (4.10
            +(3.25e5/ 783.**2)*lor(omega, 783.*cm_1, 8.*cm_1)
            +(1.04e6/1510.**2)*lor(omega,1510.*cm_1,80.*cm_1) # should be inactive
            )
    return perp, par

def permittivity_Cai(omega):
    """
    Epsilon of hexagonal boron nitride/epsilon_0.
    This is a two-component permittivity for in-plane electric field,
    out-of-plane electric field.

    This is based on Cai et al., 10.1016/j.ssc.2006.10.040 .
    """
    perp = (4.87
            +1.83*lor(omega, 1372.*cm_1, 0.)
            )
    par  = (2.95
            + 0.61*lor(omega, 746.*cm_1, 0.)
            )

    return perp, par

def permittivity_Cai_variable(omega,widthperp = 52.4, widthpar = 15.3):
    """
    Epsilon of hexagonal boron nitride/epsilon_0.
    This is a two-component permittivity for in-plane electric field,
    out-of-plane electric field.

    Optional parameters widthperp, widthpar are decay rates
    (in cm_1 -- WARNING: NON-CONSISTENT UNITS) to add losses to the
    Cai model (see permittivity_Cai) which does not specify losses.

    The default losses are made up.
    """
    warnings.warn('permittivity_Cai_variable is deprecated - WILL BE REMOVED')
    perp = (4.87
            +1.83*lor(omega, 1372.*cm_1, widthperp*cm_1)
            )
    par  = (2.95
            + 0.61*lor(omega, 746.*cm_1, widthpar*cm_1)
            )

    return perp, par

def permittivity_Cai_lossy(omega,decay_inplane=7*cm_1,decay_outplane=2*cm_1):
    """
    Epsilon of hexagonal boron nitride/epsilon_0.
    This is a two-component permittivity for in-plane electric field,
    out-of-plane electric field.

    Optional parameters decay_inplane, decay_outplane are amplitude decay
    rates (in s^-1) to add losses to the Cai model (see permittivity_Cai)
    which does not specify losses.
    Their default values are taken from permittivity_Caldwell().
    """
    perp = (4.87
            +1.83*lor(omega, 1372.*cm_1, decay_inplane)
            )
    par  = (2.95
            + 0.61*lor(omega, 746.*cm_1, decay_outplane)
            )

    return perp, par

def permittivity_Caldwell(omega):
    """
    Epsilon of hexagonal boron nitride/epsilon_0.
    This is a two-component permittivity for in-plane electric field,
    out-of-plane electric field.

    This is a "best guess" by J. Caldwell, used to produce Figure 1b in his
    paper arXiv:1404.0494.
    """
    perp = (4.90
            + 2.001*lor(omega, 1360.*cm_1, 7*cm_1)
            )
    par  = (2.95
            + 0.5262*lor(omega, 760.*cm_1, 2*cm_1)
            )

    return perp, par


def permittivity_Caldwell_isotopic(omega, isotope = ''):
    isotope_split = isotope.split(sep='_')
    switcher = {'10' : _permittivity_Caldwell_10,
                '11' : _permittivity_Caldwell_11,
                '' : _permittivity_Caldwell_mixed,
                'idealized' : _permittivity_Caldwell_idealized}
    kwargs = {}
    if len(isotope_split) > 1:
        kwargs["factor"] = float(isotope_split[1])

    return switcher[isotope_split[0]](omega,**kwargs)

def _permittivity_Caldwell_10(omega,**kwargs):
    
    perp = (5.1
            + 2.0400 * lor(omega, 1394.5*cm_1, 1.8*cm_1)
            )
    par  = (2.5
            + 0.3968 * lor(omega, 785.*cm_1, 1*cm_1)
            )

    return perp, par

def _permittivity_Caldwell_11(omega,**kwargs):
    
    perp = (5.32
            + 2.1267 * lor(omega, 1359.8*cm_1, 2.1*cm_1)
            )
    par  = (3.15
            + 0.5116 * lor(omega, 755.*cm_1, 1*cm_1)
            )
    return perp, par

def _permittivity_Caldwell_mixed(omega,**kwargs):
    
    perp = (4.90
            + 1.9049*lor(omega, 1366.2*cm_1, 7*cm_1)
            )
    par  = (2.95
            + 0.5262*lor(omega, 760.*cm_1, 2*cm_1)
            )
    return perp, par

def _permittivity_Caldwell_idealized(omega, factor):
    
    perp = (5.32
            + 2.1267 * lor(omega, 1359.8*cm_1, 2.1*cm_1*factor)
            )
    par  = (3.15
            + 0.5116 * lor(omega, 755.*cm_1, 1*cm_1*factor)
            )
    return perp, par


permittivity = permittivity_Cai


import numpy as np
from scipy.special import gamma
from mpmath import polylog
from mpmath import exp as mpexp

def unit_sphere_volume(dim):
    r'''returns \Omega_D such that the volume of a D-dimensional sphere
    of radius r is \Omega_D r^D '''
    return (np.pi**(dim/2.))/(gamma(1 + dim/2.))

def unit_sphere_surface(dim):
    r'''returns S_D such that the surface of a D-dimensional sphere
    of radius r is S_D r^(D-1) '''
    return dim * unit_sphere_volume(dim)

def fermi_dirac_int(s, x):
    r'''returns $F_s(x) = 1/\Gamma(1+s) \int_0^\infty t^s/(1+e^(t-x))dt 
    = -Li_{s+1}(-e^x)'''
    return -float(polylog(s+1, -mpexp(x)))

def average_maldague(funct, 
                     chemical_potential, 
                     temperature, 
                     sampling = None, 
                     weights = None, 
                     num = 51, 
                     quadrature = 'uniform'):
    '''
    Performs the Maldague integral
    
    $$I(f, \mu, T) = \int_{-\infty}^\infty dx/(4 \cosh^2 (x/2)) f(\mu + k_B T x)$$
    
    by approxximating it as
    
    $$I(f, \mu, T)\approx \sum_i f(\mu + k_B T *sampling_i) weights_i $$
    
    If sampling and weights are not provided they will be calculated from a 
    quadrature of the transformed integral
    $$ I(f, \mu, T) = \frac{1}{2}\int_{-1}^1 dy f(\mu + k_B T * 2*\artanh(y))$$
    using num sampling points.
    
    Available quadrature are
        - 'gauss_legendre' 
            Gauss-Legendre quadrature of order num on [-1,1]
        - 'uniform'
            num uniformely spaced points in [-1+1/num, 1-1/num], 
            i.e -1+1/num +2i/num with i = 0...num-1
    
    '''
    if (sampling is None) and (weights is None):
        if quadrature == 'gauss-legendre':
            y, w = np.polynomial.legendre.leggauss(num)
            sampling = 2 * np.arctanh(y)
            weights = 0.5 * w
        elif quadrature == 'uniform':
            y = np.linspace(-1+1./num,1-1./num,num)
            sampling = 2 * np.arctanh(y)
            weights = np.ones([num])/num
        else:
            raise ValueError('Unknown quadrature type {}'.format(quadrature))
    I = np.tensordot(weights, np.array([funct(chemical_potential + temperature * xi) for xi in sampling]), axes = 1)
    return I

# def _average_maldague(funct, chemical_potential, temperature, num = 101, sampling = None):
#     if sampling is None:
#         y, w = np.polynomial.legendre.leggauss(num)
#         x = 2*np.arctanh(y)
#         I =0.5 * np.tensordot(w, np.array([funct(chemical_potential + temperature * xi) for xi in x]), axes = 1)
#         return I
#     else:
#         chi_sample = np.array([funct(chemical_potential + y) for y in sampling])
#         factor = np.expand_dims(1./(4. * temperature * np.cosh(0.5 * sampling /temperature)**2), axis = [i for i in range(1, chi_sample.ndim)])
#         return np.trapz(chi_sample * factor, x = sampling, axis = 0)

def lor(omega, omega_r, gamma):
    """
    Lorentz shape - damped oscillator. (see source for definition)
    Used for vibrational modes in permittivities.
    """
    return (omega_r**2)/(omega_r**2 - 1.j*omega*gamma - omega**2)